/******************************************************************************
 * $Id$
 *
 * Project:  OpenGIS Simple Features Reference Implementation
 * Purpose:  Definitions related to support for use of SFCGAL in OGR.
 *           This file is only intended to be pulled in by OGR implementation
 *           code directly accessing SFCGAL.
 * Author:   Avyav Kumar Singh <avyavkumar at gmail dot com>
 *
 ******************************************************************************
 * Copyright (c) 2016, Avyav Kumar Singh <avyavkumar at gmail dot com>
 *
 * SPDX-License-Identifier: MIT
 ****************************************************************************/

#ifdef HAVE_SFCGAL

#include <SFCGAL/capi/sfcgal_c.h>

#endif
